from dataclasses import dataclass, field

from gpt_multi_atomic_agents.functions_dto import FunctionCallSchema
from .blackboard import FunctionCallBlackboard, GraphQLBlackboard, Message


@dataclass
class FunctionCallBlackboardAccessor:
    """Accessor for client to simplify integration and update of the Blackboard."""

    _blackboard: FunctionCallBlackboard = field(
        default_factory=lambda: FunctionCallBlackboard()
    )  # The inner blackboard - should only be accessed by the framework, not the client

    def get_new_functions(self) -> list[FunctionCallSchema]:
        """Get the new function calls which the client needs to execute. After executing, the client should call update_with_new_client_data() to pass in the updated and/or missing data."""
        return self._blackboard.internal_newly_generated_functions

    def get_new_messages(self) -> list[Message]:
        """Gets newly created messages that should be displayed to the user."""
        return self._blackboard.internal_newly_generated_messages

    def update_with_new_client_data(
        self, new_client_functions_representing_data: list[FunctionCallSchema]
    ) -> None:
        """Update the blackboard with the updated client data, to prepare for a new generation."""
        self._blackboard.set_user_data(user_data=new_client_functions_representing_data)


@dataclass
class MutationsAndQueries:
    previously_generated_mutation_calls: list[str] = field(default_factory=list)
    previously_generated_queries: list[str] = field(default_factory=list)


@dataclass
class GraphQLBlackboardAccessor:
    """Accessor for client to simplify integration and update of the Blackboard."""

    _blackboard: GraphQLBlackboard = field(
        default_factory=lambda: GraphQLBlackboard()
    )  # The inner blackboard - should only be accessed by the framework, not the client

    def get_new_mutations_and_queries(self) -> MutationsAndQueries:
        """Get the new mutations and queries which the client needs to execute. After executing, the client should call update_with_new_client_data() to pass in the updated and/or missing data."""
        queries = []  # TODO update when we add generation of GraphQL queries
        return MutationsAndQueries(
            previously_generated_mutation_calls=self._blackboard.internal_previously_generated_mutation_calls,
            previously_generated_queries=queries,
        )

    def get_new_messages(self) -> list[Message]:
        """Gets newly created messages that should be displayed to the user."""
        return self._blackboard.internal_newly_generated_messages

    def update_with_new_client_data(self, user_data: str) -> None:
        """Update the blackboard with the updated client data, to prepare for a new generation."""
        self._blackboard.set_user_data(user_data)


BlackboardAccessor = FunctionCallBlackboardAccessor | GraphQLBlackboardAccessor



from dataclasses import dataclass
from atomic_agents.agents.base_agent import (
    BaseIOSchema,
    BaseAgent,
    BaseAgentConfig,
)
from atomic_agents.lib.components.system_prompt_generator import SystemPromptGenerator
from pydantic import Field


from . import util_ai
from .config import Config


@dataclass
class AgentDescription:
    agent_name: str = Field(
        description="The name of the agent", examples=["Creature Creator"]
    )
    description: str = Field(
        description="The description of this agent, its purpose and capabilities."
    )
    topics: list[str] = Field(
        description="This agent ONLY generates if user mentioned one of these topics"
    )
    agent_parameters: list[str] = Field(
        description="A list of agent parameters that you can extract from the user's prompt."
    )  # Agent Parameters can be used by client to know what context to include in generation requests.


def _build_chat_agent_description(description: str) -> AgentDescription:
    return AgentDescription(
        agent_name="chat",
        description=description,
        topics=[],
        agent_parameters=["subjects"],
    )


ParamNameToValues = dict[str, list[str]]


class RecommendedAgent(BaseIOSchema):
    """
    This schema represents one agent that you recommend be used to handle the user's prompt.
    The recommendation includes the name of the agent, and a version of the user's prompt that has been rewritten to suit that agent.
    """

    agent_name: str = Field(description="The name of the agent")
    rewritten_user_prompt: str = Field(
        description="The user's prompt, rewritten to suit this agent"
    )
    agent_parameters: ParamNameToValues = Field(
        description="Agent Parameters that you extracted from the user's prompt"
    )


class AgentExecutionPlanSchema(BaseIOSchema):
    """
    This schema represents a generated plan to execute agents to fulfill the user's request. The chat message should be non-technical - do NOT mention agents.
    """

    chat_message: str = Field(
        description="The chat response to the user's message - a friendly non-technical message. Do NOT mention agents."
    )
    recommended_agents: list[RecommendedAgent] = Field(
        description="The ordered list of agents that you recommend should be used to handle the user's prompt. Only the most relevant agents should be recommended."
    )
    # TODO: could group the agents via list of ParallelAgentsGroup - this could also allow client to execute the agents in stages, allowing for HITL


class RouterAgentInputSchema(BaseIOSchema):
    """
    This schema represents the input to the Router agent.
    The schema contains the user's prompt and the list of available agents. Each agent has a special purpose. You need to recommend one or more agents to handle the users prompt.
    """

    user_prompt: str = Field(description="The chat message from the user", default="")
    agent_descriptions: list[AgentDescription] = Field(
        description="The list of available agents, describing their abilities and topics"
    )
    previous_plan: AgentExecutionPlanSchema | None = Field(
        description="The previously executed plan which the user wants you to modify",
        default=None,
    )


class RouterAgentOutputSchema(BaseIOSchema):
    """
    This schema represents the output of the Router agent.
    """

    execution_plan: AgentExecutionPlanSchema = Field(
        description="The generated plan to execute the agents"
    )


def _build_system_prompt_generator_custom() -> SystemPromptGenerator:
    return SystemPromptGenerator(
        background=[
            "You are a router bot that recommends the most suitable of the available AI agents to handle the user's prompt.",
        ],
        steps=[
            # TODO: revise/improve these steps
            "For each agent, consider whether it needs to be run to fulfull the user's prompt",
            "Only select agents that are really relevant to the user's prompt",
            "If you find no suitable agent, then generate a polite message to explain to the user that you cannot handle this request",
            "For each selected agent, rewrite the user's prompt to suit that agent",
        ],
        output_instructions=[
            "Take the user prompt and match it to a sequence of one or more of the available agents. If no suitable agent is available, then generate a polite message to explain to the user that you cannot handle this request."
        ],
    )


def create_router_agent(config: Config) -> BaseAgent:
    """
    Create a Router agent which can recommend one or more agents to handle the user's prompt. For quality it rewrites the user prompt for each agent.
    - this approach prevents agents answering prompts that are not really for them
    """
    client, model, max_tokens = util_ai.create_client(_config=config)

    agent = BaseAgent(
        config=BaseAgentConfig(
            client=client,
            model=model,
            system_prompt_generator=_build_system_prompt_generator_custom(),
            input_schema=RouterAgentInputSchema,
            output_schema=RouterAgentOutputSchema,
            max_tokens=max_tokens,
        )
    )
    return agent


def build_input(
    user_prompt: str,
    agent_descriptions: list[AgentDescription],
    chat_agent_description: str,
    previous_plan: AgentExecutionPlanSchema | None = None,
) -> RouterAgentInputSchema:
    agent_descriptions.append(_build_chat_agent_description(chat_agent_description))

    return RouterAgentInputSchema(
        user_prompt=user_prompt,
        agent_descriptions=agent_descriptions,
        previous_plan=previous_plan,
    )



import logging
import typing

from rich.console import Console

from cornsnake import util_time

from . import prompts_router
from .agent_definition import (
    AgentDefinitionBase,
)
from .config import Config
from . import util_print_agent

console = Console()

logger = logging.getLogger(__file__)


def _convert_agent_to_description(
    agent: AgentDefinitionBase,
) -> prompts_router.AgentDescription:
    return prompts_router.AgentDescription(
        agent_name=agent.agent_name,
        description=agent.description,
        topics=agent.get_topics(),
        agent_parameters=agent.get_agent_parameters(),
    )


def _convert_agents_to_descriptions(
    agents: list[AgentDefinitionBase],
) -> list[prompts_router.AgentDescription]:
    all_agents = agents
    return [_convert_agent_to_description(a) for a in all_agents]


def generate_plan(
    agent_definitions: list[AgentDefinitionBase],
    chat_agent_description: str,
    _config: Config,
    user_prompt: str,
    previous_plan: prompts_router.AgentExecutionPlanSchema | None = None,
) -> prompts_router.AgentExecutionPlanSchema:
    agent_descriptions = _convert_agents_to_descriptions(agents=agent_definitions)
    return generate_plan_via_descriptions(
        agent_descriptions=agent_descriptions,
        chat_agent_description=chat_agent_description,
        _config=_config,
        user_prompt=user_prompt,
        previous_plan=previous_plan,
    )


def generate_plan_via_descriptions(
    agent_descriptions: list[prompts_router.AgentDescription],
    chat_agent_description: str,
    _config: Config,
    user_prompt: str,
    previous_plan: prompts_router.AgentExecutionPlanSchema | None = None,
) -> prompts_router.AgentExecutionPlanSchema:
    console.log("Routing...")
    """
    Generate an agent execution plan to fulfill the user prompt, using the provided agents.
    - can be called again, with new user prompt, providing human-in-the-loop feedback.

    note: calling this router seperately from generation (agent execution) helps to reduce the *perceived* time taken to generate, since the user gets an (intermediate) response earlier.
    """

    start = util_time.start_timer()

    # TODO: make router reject irrelevant user prompts (if no matching agents + it does not fit the chat_agent_description)

    # TODO: optimizate router:
    # - possibly run it on smaller (and faster) LLM
    # - could allow for Classifier based router, but then cannot rewrite prompts
    router_agent = prompts_router.create_router_agent(config=_config)
    response = typing.cast(
        prompts_router.RouterAgentOutputSchema,
        router_agent.run(
            prompts_router.build_input(
                user_prompt=user_prompt,
                agent_descriptions=agent_descriptions,
                chat_agent_description=chat_agent_description,
                previous_plan=previous_plan,
            )
        ),
    )

    util_print_agent.print_router_assistant(response, _config=_config)
    time_taken = util_time.end_timer(start=start)
    console.log(f"  time taken: {util_time.describe_elapsed_seconds(time_taken)}")

    return response.execution_plan



import typing

from atomic_agents.agents.base_agent import (
    BaseIOSchema,
)

from rich.console import Console
from rich.text import Text

from . import prompts_router, config
from .agent_definition import (
    AgentDefinitionBase,
    FunctionAgentDefinition,
    GraphQLAgentDefinition,
)
from .functions_dto import FunctionAgentOutputSchema
from .graphql_dto import GraphQLAgentOutputSchema

console = Console()


def print_agent(
    agent: prompts_router.RecommendedAgent,
    _config: config.Config,
    max_prompt_out_len: int = 200,
    prefix="",
) -> None:
    rewritten_user_prompt = agent.rewritten_user_prompt
    if not _config.is_debug and len(rewritten_user_prompt) > max_prompt_out_len:
        rewritten_user_prompt = agent.rewritten_user_prompt[:max_prompt_out_len] + "..."
    console.print(Text(f" {prefix} - [{agent.agent_name}]", style="cyan"))
    console.print(Text(f"  <-- '{rewritten_user_prompt}'", style="blue"))


def print_router_assistant(
    message: prompts_router.RouterAgentOutputSchema, _config: config.Config
) -> None:
    console.print(
        f":robot: [bold cyan]Assistant [router]: {message.execution_plan.chat_message}[/bold cyan]"
    )
    console.print(Text("  - recommended agents:", style="blue"))
    for agent in message.execution_plan.recommended_agents:
        print_agent(agent, _config=_config)


def print_assistant_message_only(chat_message: str, agent_name="general"):
    console.print(
        f":robot: [bold green]Assistant [{agent_name}]: {chat_message}[/bold green]"
    )


def _print_assistant_base(chat_message: str, output: typing.Any, agent_name="general"):
    print_assistant_message_only(chat_message=chat_message, agent_name=agent_name)
    console.print(Text("  New calls:", style="yellow"))
    console.print(output)


def print_assistant_message(message: str):
    initial_message = FunctionAgentOutputSchema(
        chat_message=message, generated_function_calls=[]
    )
    print_assistant_functions(message=initial_message)


def print_assistant_functions(message: FunctionAgentOutputSchema, agent_name="general"):
    return _print_assistant_base(
        message.chat_message, message.generated_function_calls, agent_name=agent_name
    )


def _print_assistant_graphql(message: GraphQLAgentOutputSchema, agent_name="general"):
    return _print_assistant_base(
        message.chat_message, message.generated_mutations, agent_name=agent_name
    )


def print_assistant_output(
    response: BaseIOSchema, agent_definition: AgentDefinitionBase
) -> None:
    if isinstance(agent_definition, FunctionAgentDefinition):
        return print_assistant_functions(response, agent_definition.agent_name)
    elif isinstance(agent_definition, GraphQLAgentDefinition):
        return _print_assistant_graphql(response, agent_definition.agent_name)
    else:
        raise RuntimeError("Not a recognised AgentDefinitionBase")


def print_user_prompt(user_prompt: str) -> None:
    console.print(f":sunglasses: You: {user_prompt}")



from pydantic import Field
from gpt_multi_atomic_agents.util_pydantic import CustomBaseModel
from .functions_dto import (
    FunctionCallSchema,
    FunctionSpecSchema,
    ParameterSpec,
    ParameterType,
)

creature_agent_name = "Creature Creator"


class FunctionAgentDefinitionMinimal(CustomBaseModel):
    agent_name: str = Field(
        description="The name of the agent", examples=[creature_agent_name]
    )
    description: str = Field(
        description="The description of this agent, its purpose and capabilities."
    )
    accepted_functions: list[FunctionSpecSchema] = Field(
        description="The set of 'input' function calls that this agent understands. Each agent should understand its own output, but can also understand a subset of the output of other agents. This allows the agents to collaborate."
    )
    functions_allowed_to_generate: list[FunctionSpecSchema] = Field(
        description="The set of 'output' function calls that this agent generates."
    )
    topics: list[str] = Field(
        description="This agent ONLY generates if user mentioned one of these topics"
    )


# Build examples programatically = less error prone
creature_icons = ["sheep-icon", "wolf-icon", "grass-icon", "human-icon", "other-icon"]
terrain_types = ["mountain", "marsh", "prairie", "coast", "water"]

function_create_creature = FunctionSpecSchema(
    function_name="AddCreature",
    description="Adds a new creature to the world (not vegetation)",
    parameters=[
        ParameterSpec(name="creature_name", type=ParameterType.string),
        ParameterSpec(
            name="allowed_terrain",
            type=ParameterType.string,
            allowed_values=terrain_types,
        ),
        ParameterSpec(name="age", type=ParameterType.int),
        ParameterSpec(
            name="icon_name", type=ParameterType.string, allowed_values=creature_icons
        ),
    ],
)

example_creeature_creator_agent = FunctionAgentDefinitionMinimal(
    agent_name=creature_agent_name,
    description="Creates new creatures given the user prompt. Ensures that ALL creatures mentioned by the user are created.",
    accepted_functions=[function_create_creature],
    functions_allowed_to_generate=[function_create_creature],
    topics=["creature", "summary"],
)

example_create_creature_call__wolf = FunctionCallSchema(
    agent_name=creature_agent_name,
    function_name=function_create_creature.function_name,
    parameters={
        "creature_name": "wolf",
        "allowed_terrain": "mountain",
        "age": "5",
        "icon_name": "wolf-icon",
    },
)



import logging
import os
from cornsnake import util_toml
from dataclasses import dataclass
from enum import StrEnum, auto

logger = logging.getLogger(__file__)


class AI_PLATFORM_Enum(StrEnum):
    openai = auto()
    groq = auto()
    bedrock_anthropic = auto()


GROQ_MODEL = "llama-3.1-70b-versatile"  # 'llama-3.1-70b-versatile' #"llama-3.1-8b-instant"  #  llama3-8b-8192

OPEN_AI_MODEL = "gpt-4o"  # "gpt-3.5-turbo"

ANTHROPIC_MODEL = "anthropic.claude-3-5-sonnet-20240620-v1:0"
ANTHROPIC_MAX_TOKENS = 8192


@dataclass
class Config:
    ai_platform: AI_PLATFORM_Enum = AI_PLATFORM_Enum.bedrock_anthropic
    model: str = ANTHROPIC_MODEL
    max_tokens: int = ANTHROPIC_MAX_TOKENS
    is_debug: bool = False
    delay_between_calls_in_seconds: float = 0.0
    temp_data_dir_path: str = "data-generated"


def _get_path_to_ini(path_to_ini: str) -> str:
    path_to_ini = path_to_ini
    if not os.path.exists(path_to_ini):
        # look in current directory
        path_to_ini = os.path.join(os.getcwd(), path_to_ini)
    logger.info(f"Reading config from '{path_to_ini}'")
    return path_to_ini


def load_config(path_to_ini: str) -> Config:
    path_to_file = _get_path_to_ini(path_to_ini)
    config = Config
    util_toml.read_config_ini_file(path_to_file=path_to_file, config_object=config)
    return config



# A generic Agent prompt.
# - more specialized prompts can be set when creating the relevant AgentSpec.
from . import util_output
from .config import Config


GENERIC_AGENT_PROMPT_TEMPLATE = """
Examine the provided user prompt, the available functions and the previously generated functions. The user's request should be handled by generating function calls.
ONLY generate if the user asked about one of these topics: {TOPICS}.

AVAILABLE_FUNCTIONS: ```{AVAILABLE_FUNCTIONS}```

For each available function, do the following:
- check is the function relevant to the user's prompt
- check what functions have already been called, that would setup the necessary state
- only output the function if it is relevant and necessary
- only output the function if all of its parameters have values
- output the functions in the best possible order

You must perform all your reasoning and analysis within a single set of <thinking> tags. Use the following structure:

<thinking>
For each generated function:
[function-name]: [function-name]
  - name: [the name of the function]
  - parameters: [the parameters and their values]
[Continue for all items]


Overall analysis:
- [Any general observations or patterns noticed across function calls]
- [Any potential relationships or dependencies between function calls]

</thinking>

After your thinking process, output the generated function calls.

Notes:
- Ensure the functions are called in the best possible order
- For each function call, ensure that all required parameters are provided.
- For each function call, ensure that all interesting information is included.
- Remove any function calls that are not for you to generate - see your 'AVAILABLE_FUNCTIONS' list
- Only generate function calls if really necessary - it is OK to output with no functions.
"""


def build_agent_prompt(
    allowed_functions_to_generate_names: list[str], topics: list[str], _config: Config
) -> str:
    def _join(strings: list[str]):
        return ", ".join(strings)

    prompt = GENERIC_AGENT_PROMPT_TEMPLATE.replace("{TOPICS}", _join(topics)).replace(
        "{AVAILABLE_FUNCTIONS}", _join(allowed_functions_to_generate_names)
    )

    util_output.print_debug(f"prompt: {prompt}", config=_config)

    return prompt



# A generic Agent prompt.
# TODO: more specialized prompts can be set when creating the relevant AgentDefinition.


GENERIC_AGENT_PROMPT_TEMPLATE = """
Examine the provided user prompt, the available GraphQL mutation definitions, the graphql_input data and the previously generated GraphQL mutations. The user's request should be handled by generating GraphQL mutations.
ONLY generate if the user asked about one of the agent topics.

For each available mutation, do the following:
- check is the mutation relevant to the user's prompt
- check the existing graphql_data and what mutations have already been called, that would setup the necessary state
- only output the mutation if it is relevant and necessary
- only output the mutation if all of its parameters have values
- output the mutations in the best possible order

You must perform all your reasoning and analysis within a single set of <thinking> tags. Use the following structure:

<thinking>
For each generated mutation:
[mutation-name]: [mutation-name]
  - name: [the name of the mutation]
  - parameters: [the parameters and their values]
[Continue for all items]


Overall analysis:
- [Any general observations or patterns noticed across mutation calls]
- [Any potential relationships or dependencies between mutation calls and also the existing GraphQL data]

</thinking>

After your thinking process, output only the necessary generated mutation calls.

Notes:
- Ensure the mutations are called in the best possible order
- For each mutation call, ensure that all required parameters are provided.
- For each mutation call, ensure that all interesting information is included.
- Remove any mutation calls that are not for you to generate - see your 'AVAILABLE_MUTATIONS'.
- Only generate mutations if really necessary - check against the input graphql_data. It is OK to output with no mutations.
"""


def build_agent_prompt() -> str:
    return GENERIC_AGENT_PROMPT_TEMPLATE



from enum import StrEnum
from atomic_agents.agents.base_agent import (
    BaseIOSchema,
)
from pydantic import Field


class ParameterType(StrEnum):
    """Defines the type of a parameter."""

    int = "int"
    float = "float"
    string = "string"
    date = "datetime"
    # TODO: add more parameter types


class ParameterSpec(BaseIOSchema):
    """Defines one parameter of a function, including its name, type and allowed values. If allowed values are empty, then values must match the 'type'."""

    name: str
    type: ParameterType
    allowed_values: list[str] = Field(default_factory=lambda: [])


class FunctionSpecSchema(BaseIOSchema):
    """This schema represents the definition of a function call that can be generated."""

    description: str = Field(description="Describes what the function does")
    function_name: str = Field(description="The name of the function")
    parameters: list[ParameterSpec] = Field(
        description="Named parameters of the function"
    )


class FunctionCallSchema(BaseIOSchema):
    """This schema represents a function call that was already generated."""

    agent_name: str = Field(
        description="The name of the agent that generated the function call"
    )
    function_name: str = Field(description="The name of the function")
    parameters: dict[str, str] = Field(
        description="The named parameters and their values"
    )


class FunctionAgentInputSchema(BaseIOSchema):
    """
    This schema represents the input to the agent.
    The schema contains previously generated function calls, and the list of allowed generated functions.
    """

    user_input: str = Field(description="The chat message from the user", default="")
    functions_allowed_to_generate: list[FunctionSpecSchema] = Field(
        description="Definitions of the functions that this agent can generate"
    )
    previously_generated_functions: list[FunctionCallSchema] = Field(
        description="Previously generated functions in this chat (some are from other agents)",
        default_factory=lambda: list,
    )


class FunctionAgentOutputSchema(BaseIOSchema):
    """
    This schema represents the output of the agent. The chat message should be non-technical - do NOT mention functions.
    """

    chat_message: str = Field(
        description="The chat response to the user's message - a friendly, non-technical message. Do NOT mention functions."
    )
    generated_function_calls: list[FunctionCallSchema] = Field(
        description="The set of new generated function calls"
    )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from atomic_agents.lib.components.system_prompt_generator import SystemPromptGenerator

from . import prompts_agent_functions, prompts_agent_graphql
from .config import Config
from .functions_dto import FunctionSpecSchema


@dataclass
class SystemPromptBuilderBase(ABC):
    _config: Config

    @abstractmethod
    def build_system_prompt(self) -> str:
        raise NotImplementedError


@dataclass
class FunctionSystemPromptBuilder(SystemPromptBuilderBase):
    allowed_functions_to_generate: list[FunctionSpecSchema]
    topics: list[str]

    def build_system_prompt(self) -> str:
        allowed_functions_to_generate_names = [
            f.function_name for f in self.allowed_functions_to_generate
        ]

        return SystemPromptGenerator(
            background=[
                "You are a helpful assistant that can only generate function calls using the provided definitions."
            ],
            steps=[
                # TODO: could break the prompt down into steps
                prompts_agent_functions.build_agent_prompt(
                    allowed_functions_to_generate_names,
                    topics=self.topics,
                    _config=self._config,
                )
            ],
            output_instructions=[
                "Your output should always be a set of zero or more generated functions, using only the allowed function definitions."
            ],
        )


@dataclass
class GraphQLSystemPromptBuilder(SystemPromptBuilderBase):
    def build_system_prompt(self) -> str:
        return SystemPromptGenerator(
            background=[
                "You are a helpful assistant that can only generates GraphQL mutations using the provided definitions."
            ],
            steps=[
                # TODO: could break the prompt down into steps
                prompts_agent_graphql.build_agent_prompt()
            ],
            output_instructions=[
                # TODO: add also output GraphQL queries
                "Your output should always be a set of zero or more GraphQL mutation calls, using only the allowed GraphQL mutations definitions."
            ],
        )



from abc import abstractmethod
import typing
from atomic_agents.agents.base_agent import (
    BaseIOSchema,
)
from pydantic import Field

from .graphql_dto import GraphQLAgentInputSchema, GraphQLAgentOutputSchema

from .system_prompt_builders import (
    FunctionSystemPromptBuilder,
    GraphQLSystemPromptBuilder,
    SystemPromptBuilderBase,
)
from .util_pydantic import CustomBaseModel

from . import util_output
from .blackboard import (
    Blackboard,
    FunctionCallBlackboard,
    GraphQLBlackboard,
    Message,
    MessageRole,
)
from .config import Config

from .functions_dto import (
    FunctionAgentInputSchema,
    FunctionAgentOutputSchema,
    FunctionSpecSchema,
)


class AgentDefinitionBase(CustomBaseModel):
    """
    Defines one function-based agent. NOT for direct use with LLM.
    """

    agent_name: str = Field(
        description="The name of the agent.", examples=["Creature Creator"]
    )
    description: str = Field(
        description="Describes the function of the agent. This acts as a mini prompt for the LLM."
    )
    input_schema: type[BaseIOSchema]
    initial_input: BaseIOSchema
    output_schema: type[BaseIOSchema]
    # TODO: could add custom prompt/prompt-extension if needed

    @abstractmethod
    def build_input(
        self, rewritten_user_prompt: str, blackboard: Blackboard, config: Config
    ) -> BaseIOSchema:
        raise NotImplementedError

    @abstractmethod
    def get_system_prompt_builder(self, _config: Config) -> SystemPromptBuilderBase:
        raise NotImplementedError

    @abstractmethod
    def update_blackboard(self, response: BaseIOSchema, blackboard: Blackboard) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_topics(self) -> list[str]:
        raise NotImplementedError

    @abstractmethod
    def get_agent_parameters(self) -> list[str]:
        raise NotImplementedError


class FunctionAgentDefinition(AgentDefinitionBase):
    input_schema: type[FunctionAgentInputSchema]
    initial_input: FunctionAgentInputSchema
    output_schema: type[FunctionAgentOutputSchema]
    accepted_functions: list[FunctionSpecSchema]
    topics: list[str]  # The agent ONLY generates if user mentioned one of these topics
    agent_parameters: list[
        str
    ]  # Optional parameters that router can try to extract from user prompt

    def _cast_blackboard(self, blackboard: Blackboard) -> FunctionCallBlackboard:
        if not isinstance(blackboard, FunctionCallBlackboard):
            raise RuntimeError("Expected blackboard to be a FunctionCallBlackboard")
        return typing.cast(FunctionCallBlackboard, blackboard)

    def _cast_response(self, response: BaseIOSchema) -> FunctionAgentOutputSchema:
        if not isinstance(response, FunctionAgentOutputSchema):
            raise RuntimeError("Expected response to be a FunctionAgentOutputSchema")
        return typing.cast(FunctionAgentOutputSchema, response)

    def build_input(
        self, rewritten_user_prompt: str, blackboard: Blackboard, config: Config
    ) -> BaseIOSchema:
        function_blackboard = self._cast_blackboard(blackboard)

        initial_input = self.initial_input
        initial_input.user_input = rewritten_user_prompt

        initial_input.previously_generated_functions = (
            function_blackboard.get_generated_functions_matching(
                self.get_accepted_function_names()
            )
        )
        util_output.print_debug(
            f"[{self.agent_name}] Previously generated funs: {initial_input.previously_generated_functions}",
            config,
        )

        return initial_input

    def get_accepted_function_names(self) -> list[str]:
        return [f.function_name for f in self.accepted_functions]

    def get_topics(self) -> list[str]:
        return self.topics

    def get_agent_parameters(self) -> list[str]:
        return self.agent_parameters

    def get_system_prompt_builder(self, _config: Config) -> SystemPromptBuilderBase:
        return FunctionSystemPromptBuilder(
            topics=self.topics,
            _config=_config,
            allowed_functions_to_generate=self.initial_input.functions_allowed_to_generate,
        )

    def update_blackboard(self, response: BaseIOSchema, blackboard: Blackboard) -> None:
        function_response = self._cast_response(response)
        function_blackboard = self._cast_blackboard(blackboard)
        function_blackboard.add_mesage(
            Message(role=MessageRole.assistant, message=function_response.chat_message)
        )
        function_blackboard.add_generated_functions(
            function_response.generated_function_calls
        )


def build_function_agent_definition(
    agent_name: str,
    description: str,
    accepted_functions: list[FunctionSpecSchema],
    functions_allowed_to_generate: list[FunctionSpecSchema],
    topics: list[str],
    agent_parameters: list[str] | None = None,
) -> FunctionAgentDefinition:
    if not agent_parameters:
        agent_parameters = []
    return FunctionAgentDefinition(
        agent_name=agent_name,
        description=description,
        accepted_functions=accepted_functions,
        input_schema=FunctionAgentInputSchema,
        initial_input=FunctionAgentInputSchema(
            functions_allowed_to_generate=functions_allowed_to_generate,
            previously_generated_functions=[],
        ),
        output_schema=FunctionAgentOutputSchema,
        topics=topics,
        agent_parameters=agent_parameters,
    )


class GraphQLAgentDefinition(AgentDefinitionBase):
    input_schema: type[GraphQLAgentInputSchema]
    initial_input: GraphQLAgentInputSchema
    output_schema: type[GraphQLAgentOutputSchema]

    def _cast_blackboard(self, blackboard: Blackboard) -> GraphQLBlackboard:
        if not isinstance(blackboard, GraphQLBlackboard):
            raise RuntimeError("Expected blackboard to be a GraphQLBlackboard")
        graphql_blackboard = typing.cast(GraphQLBlackboard, blackboard)
        return graphql_blackboard

    def _cast_response(self, response: BaseIOSchema) -> GraphQLAgentOutputSchema:
        if not isinstance(response, GraphQLAgentOutputSchema):
            raise RuntimeError("Expected response to be a GraphQLAgentOutputSchema")
        return typing.cast(GraphQLAgentOutputSchema, response)

    def build_input(
        self, rewritten_user_prompt: str, blackboard: Blackboard, config: Config
    ) -> BaseIOSchema:
        graphql_blackboard = self._cast_blackboard(blackboard)

        initial_input = self.initial_input

        initial_input.user_input = rewritten_user_prompt
        initial_input.graphql_data = graphql_blackboard.get_user_data()

        initial_input.previously_generated_mutations = (
            graphql_blackboard.get_generated_mutations_matching(
                initial_input.accepted_graphql_schemas
            )
        )
        util_output.print_debug(
            f"[{self.agent_name}] build_input(): {initial_input}", config
        )
        util_output.print_debug(
            f"[{self.agent_name}] Matching previously generated mutations: {initial_input.previously_generated_mutations}",
            config,
        )

        return initial_input

    def get_system_prompt_builder(self, _config: Config) -> SystemPromptBuilderBase:
        return GraphQLSystemPromptBuilder(_config=_config)

    def get_topics(self) -> list[str]:
        return self.initial_input.topics

    def get_agent_parameters(self) -> list[str]:
        return self.initial_input.agent_parameters

    def update_blackboard(self, response: BaseIOSchema, blackboard: Blackboard) -> None:
        graphql_blackboard = self._cast_blackboard(blackboard)
        graphql_response = self._cast_response(response)

        graphql_blackboard.add_generated_mutations(graphql_response.generated_mutations)
        graphql_blackboard.add_mesage(
            Message(role=MessageRole.assistant, message=graphql_response.chat_message)
        )


def build_graphql_agent_definition(
    agent_name: str,
    description: str,
    accepted_graphql_schemas: list[str],
    mutations_allowed_to_generate: list[str],
    topics: list[str],
    agent_parameters: list[str] | None = None,
) -> GraphQLAgentDefinition:
    if not agent_parameters:
        agent_parameters = []
    return GraphQLAgentDefinition(
        agent_name=agent_name,
        description=description,
        input_schema=GraphQLAgentInputSchema,
        initial_input=GraphQLAgentInputSchema(
            accepted_graphql_schemas=accepted_graphql_schemas,
            graphql_data="",
            mutations_allowed_to_generate=mutations_allowed_to_generate,
            previously_generated_mutations=[],
            topics=topics,
            agent_parameters=agent_parameters,
        ),
        output_schema=GraphQLAgentOutputSchema,
    )
